﻿using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;

namespace Vam_preset_maker
{


    public class PresetMaker : MVRScript
    {
        UIDynamicButton _processButton;


        UIDynamicToggle SaveAppearance;
        UIDynamicToggle SaveClothing;
        UIDynamicToggle SaveSkin;
        UIDynamicToggle SaveHair;

        JSONStorableBool SaveAppearanceBool = new JSONStorableBool("AppearancePresets", true);
        JSONStorableBool SaveClothingBool = new JSONStorableBool("ClothingPresets", true);
        JSONStorableBool SaveSkinBool = new JSONStorableBool("SkinPresets", true);
        JSONStorableBool SaveHairBool = new JSONStorableBool("HairPresets", true);
        List<JSONStorableBool> _saveOptions;

        DAZCharacterSelector _dazCharacterSelector;

        CamPositions camPositions = new CamPositions();

        bool _waitingOnScreenShot = false;
        bool firstScreenshot = true;
        bool StopRequested = false;

        string fileName { get; set; }
        string imageName { get; set; }
        string _presetPath { get; set; } = string.Empty;


        CacheSystem cacheSystem;

        JSONStorableStringChooser jssc;



        public override void Init()
        {
            //create folders
            if (!FileManagerSecure.DirectoryExists("Custom/PluginData/Chokaphi/PresetMaker/"))
            {
                FileManagerSecure.CreateDirectory("Custom/PluginData/Chokaphi/PresetMaker/");
            }

            SaveAppearance = CreateToggle(SaveAppearanceBool);
            SaveClothing = CreateToggle(SaveClothingBool);
            SaveSkin = CreateToggle(SaveSkinBool);
            SaveHair = CreateToggle(SaveHairBool);

            _saveOptions = new List<JSONStorableBool>() { SaveAppearanceBool, SaveClothingBool, SaveHairBool, SaveSkinBool };

            List<string> CreatorsList = new List<string>() { "All" };
            List<string> varfiles = GetFiles("AddonPackages", "*.var").ToList();
            foreach (string path in varfiles)
            {
                //string varName = FileManagerSecure.NormalizePath(path);
                string varName = FileManagerSecure.GetFileName(path);

                string[] varNames = varName.Split('.');
                if (!CreatorsList.Contains(varNames[0]))
                {
                    CreatorsList.Add(varNames[0]);
                }

            }

            jssc = new JSONStorableStringChooser("Creators", CreatorsList, "All", "Select Creator to filter by");
            UIDynamicPopup pop = CreateScrollablePopup(jssc, true);


            cacheSystem = new CacheSystem(this);


            camPositions.StoreCamPosition(null, SuperController.singleton.MonitorRig.transform);
 

            _processButton = CreateButton("process Vars", true);

            if (base.containingAtom.mainController == null)
            {
                SuperController.LogError("Please add this plugin to a PERSON atom.");
                return;
            }

            _dazCharacterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();


            _processButton.button.onClick.AddListener(() =>
            {


                SuperController.LogMessage("Processing VAR Files");
                _processButton.button.enabled = false;

                StartCoroutine(ProcessLooks("AddonPackages", "*.json"));

            });

        }

        IEnumerator StoreCameraPostions()
        {
            Text help = SuperController.singleton.screenshotPreview.GetComponentsInChildren<Text>().Where(x => x.name == "HelpText").FirstOrDefault();
            camPositions.OrigHelpText = help.text;
            camPositions.OrigHelpColor = help.color;

            if (SaveAppearanceBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/AppearancePresets", RecordCamPos);
                help.text = "Set Camera you want to use for Appearance Preset";
                help.color = Color.cyan;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveClothingBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/ClothingPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Clothing Preset";
                help.color = Color.yellow;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveSkinBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/SkinPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Skin Preset";
                help.color = Color.red;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveHairBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/HairPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Hair Preset";
                help.color = Color.green;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            help.text = camPositions.OrigHelpText;
            help.color = camPositions.OrigHelpColor;
        }

        void RecordCamPos(string presetType)
        {
            presetType = FileManagerSecure.GetFileName(presetType);
            camPositions.StoreCamPosition(presetType, SuperController.singleton.screenshotCamera.transform);

            _waitingOnScreenShot = false;
        }

        string PresetTypePath(string presetType)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    return "Custom/Atom/Person/Appearance/";

                case "ClothingPresets":
                    return "Custom/Atom/Person/Clothing/";

                case "SkinPresets":
                    return "Custom/Atom/Person/Skin/";

                case "HairPresets":
                    return "Custom/Atom/Person/Hair/";
            }

            return null;
        }


        string SavePreset(string presetType = "ClothingPresets")
        {
            //turn off auto expressions,blink, jaw collision
            containingAtom.GetStorableByID("EyelidControl").SetBoolParamValue("blinkEnabled", false);
            containingAtom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", false);
            containingAtom.GetStorableByID("TongueControl").SetBoolParamValue("tongueCollision", false);
            //Turn on Look at Camera


            List<MeshVR.PresetManagerControl> pmcs = containingAtom.presetManagerControls;
            MeshVR.PresetManagerControl presetControl = pmcs.Where(x => x.name == presetType).FirstOrDefault();
            //set preset name
            presetControl.SetStringParamValue("presetName", _presetPath + fileName);
            //store preset
            presetControl.CallAction("StorePreset");
            //create screenshotname
            string path = string.Concat(new string[]
                        {
                                PresetTypePath(presetType),
                                 _presetPath,
                                 imageName
                        });

            //now call the screenshot with callback
            _waitingOnScreenShot = true;

            return path;
        }

        void ScreenShotFinished(string text)
        {
            _waitingOnScreenShot = false;
        }

        void CreateFileName(string varName, string name = "Person", bool isMale = false)
        {
            string sex = "Female";
            if (isMale)
                sex = "Male";

            string[] varNames = varName.Split('.');

            _presetPath = string.Concat(new string[]
                {
                    "/",
                    "PMA-Test",




                    "/",
                    varNames[0],
                    "/",
                    //varNames[1],
                    //"/",
                });

            if (name.ToLower().Contains("person"))
            {
                name = varNames[1];
            }


            this.fileName = name + "-" + sex;


            int count = 1;
            string fileName = this.fileName;
            string testPath = "Custom/Atom/Person/Appearance" + _presetPath + "Preset_" + fileName + ".vap";
            bool fileExists = FileManagerSecure.FileExists(testPath);

            while (FileManagerSecure.FileExists(testPath))
            {
                testPath = testPath = "Custom/Atom/Person/Appearance" + _presetPath + "Preset_" + fileName + count + ".vap";

                fileExists = FileManagerSecure.FileExists(testPath);
                this.fileName = fileName + count;

                count++;
            }

            imageName = "Preset_" + this.fileName;
        }

        #region ProcessLooks helpers
        JSONNode LoadJsonString(string file)
        {
            string jsonString = FileManagerSecure.ReadAllText(file);
            //replace all SELF with var path just to load the person node with internal VAR ref
            string varName = FileManagerSecure.NormalizePath(file);
            varName = Regex.Replace(varName, "(.*)(:.+)", "$1");
            jsonString = Regex.Replace(jsonString, "(SELF)(:.+)", varName + "$2");

            return JSON.Parse(jsonString);
        }

        void LoadPersonNode(JSONNode personNode)
        {
            JSONClass asObject = personNode.AsObject;
            //load person file
            containingAtom.PreRestore(false, true);
            containingAtom.Restore(asObject, false, true, false, null, false, false, true, false);
            containingAtom.LateRestore(asObject, false, true, false, false, true, false);
            containingAtom.PostRestore(false, true);

            if (SuperController.singleton != null && asObject["id"] != null)
            {
                SuperController.singleton.RenameAtom(containingAtom, asObject["id"]);
                containingAtom.ResetPhysics(false, true);
            }
        }
        #endregion

        bool filter(string filePath)
        {
            string creatorName = Regex.Replace(filePath, @".*\\(.*)(\..*\..*\..*:)(.+)", "$1");

            if (jssc.val == "All")
                return false;

            if (creatorName == jssc.val)
                return false;

            return true;
        }

        IEnumerator ProcessLooks(string path, string pattern)
        {
            int fileCount = 0;
            int lookCount = 0;
            foreach (string filePath in GetFiles(path, pattern))
            {

                if (filter(filePath))
                {

                }
                else if (!cacheSystem.FileInCache(filePath))
                {
                    JSONNode jn = LoadJsonString(filePath);

                    if (jn["atoms"] != null)
                    {
                        fileCount++;
                        JSONNode[] personNodes = jn["atoms"].Childs.Where(x => x["type"].Value == "Person").ToArray();

                        int nodeIndex = 0;
                        foreach (JSONNode personNode in personNodes)
                        {


                            lookCount++;

                            LoadPersonNode(personNode);

                            DAZCharacterRun dazCharacterRun = _dazCharacterSelector.GetComponent<DAZCharacterRun>();
                            DAZCharacter _dazCharacter = _dazCharacterSelector.selectedCharacter;

                            yield return new WaitUntil(() => _dazCharacter.ready);
                            //wait for physics to finish
                            yield return new WaitWhile(() => dazCharacterRun.IsSimulationResetting());
                            //wait for clothing and any texture loads are finished
                            yield return new WaitWhile(() => SuperController.singleton.loadingIcon.gameObject.activeSelf);
                            //wait 2 seconds for physics to settle
                            yield return new WaitForSeconds(2);


                            string varName = FileManagerSecure.NormalizePath(filePath);
                            varName = Regex.Replace(varName, "(.*)(:.+)", "$1");
                            CreateFileName(varName, personNode["id"], _dazCharacter.isMale);


                            //check if we have already saved this look
                            //ZERO low value morphs micro values in on versions will cause duplicates any value < 0.01
                            try
                            {
                                _dazCharacterSelector.morphsControlFemaleUI.ZeroNearZero();
                                _dazCharacterSelector.morphsControlMaleUI.ZeroNearZero();
                            }
                            catch
                            {
                            }

                            //for the first screenshot of session allow user to position camera for screenshots
                            if (firstScreenshot)
                            {
                                firstScreenshot = false;
                                //record cam position vs head position
                                Transform headT = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault().transform;
                                camPositions.SetStartPosition(headT);

                                yield return StartCoroutine(StoreCameraPostions());
                            }

                            //now store screenshots based on selections
                            foreach (JSONStorableBool jsonBool in _saveOptions)
                            {
                                //if we have a match in cache skip
                                string savepath;
                                FakeStoreStorables fakeStoreStorables = new FakeStoreStorables(this);
                                string value = fakeStoreStorables.GetStorablesString(containingAtom, jsonBool.name);
                             
                                if (jsonBool.val && !cacheSystem.StorableInCache(filePath, nodeIndex, personNodes.Length, jsonBool.name, value))
                                {

                                    savepath = SavePreset(jsonBool.name);

                                    //now store the correct CAM Screenshot
                                    SuperController.singleton.DoSaveScreenshot(savepath + ".jpg", ScreenShotFinished);

                                    //adjust cam position based on Character height
                                    Transform headT = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault().transform;
                                    camPositions.UpdateHeadDeltas(headT);
                                    camPositions.LoadCamPosition(jsonBool.name, SuperController.singleton.MonitorRig.transform);


                                    //take two frames for camera to reposition.
                                    yield return new WaitForEndOfFrame();
                                    yield return new WaitForEndOfFrame();
                                    //yield return new WaitUntil(() => Input.GetKey(KeyCode.DownArrow));

                                    SuperController.singleton.SetLeftSelect();
                                    yield return new WaitWhile(() => _waitingOnScreenShot);
                                }

                            }
                            //reset cam for next load. Not needed but makes it look better if watching
                            camPositions.LoadCamPosition(null, SuperController.singleton.MonitorRig.transform);

                            //allow user to cancel coroutine
                            if (StopRequested || Input.GetKey("space"))
                            {
                                SuperController.LogMessage("Preset processing has been Stopped");
                                _processButton.button.enabled = true;
                                StopRequested = false;
                                yield break;
                            }

                            nodeIndex++;
                        }
                        if (StopRequested || Input.GetKey("space"))
                        {
                            SuperController.LogMessage("Preset processing has been Stopped");
                            _processButton.button.enabled = true;
                            StopRequested = false;
                            yield break;
                        }

                    }
                }

            }

            yield return new WaitForEndOfFrame();
            _processButton.button.enabled = true;
            //reset cam to defalut.
            camPositions.LoadCamPosition(null, SuperController.singleton.MonitorRig.transform);

            SuperController.LogMessage("Finished Proccessing all VAR json files. Files processed = " + fileCount + " looks processed = " + lookCount);
        }

        void Update()
        {
            if (!StopRequested && Input.GetKeyDown("space"))
            {
                SuperController.LogMessage("Preset processing has been requested to stop");
                StopRequested = true;
            }
        }

        static IEnumerable<string> GetFiles(string path, string pattern)
        {
            Queue<string> queue = new Queue<string>();
            queue.Enqueue(path);
            while (queue.Count > 0)
            {
                path = queue.Dequeue();
                try
                {
                    foreach (string subDir in FileManagerSecure.GetDirectories(path))
                    {
                        queue.Enqueue(subDir);
                    }
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                string[] files = null;
                try
                {
                    files = FileManagerSecure.GetFiles(path, pattern);
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                if (files != null)
                {
                    for (int i = 0; i < files.Length; i++)
                    {
                        yield return files[i];
                    }
                }
            }
        }


        #region


        #endregion

    }

    class CacheSystem
    {
        JSONClass cacheFileJSON = new JSONClass();
        readonly string Cachepath = "Custom/PluginData/Chokaphi/PresetMaker/Cache.json";
        MVRScript mVR;

        public CacheSystem(MVRScript mVR)
        {
            this.mVR = mVR;
            LoadCache();
        }

        public bool FileInCache(string filePath)
        {
            if (cacheFileJSON[filePath] == null)
            {
                return false;
            }

            if (cacheFileJSON[filePath]["nodes"].Count == cacheFileJSON[filePath]["totalNodes"].AsInt)
            {
                //if total nodes match then have stored all presetTypes for this file
                foreach (JSONClass j in cacheFileJSON[filePath]["nodes"].AsArray)
                {
                    if (j.HasKey("AppearancePresets") && j.HasKey("ClothingPresets") && j.HasKey("SkinPresets") && j.HasKey("HairPresets"))
                    {
                        //do nothing
                    }
                    else
                    {
                        return false;
                    }
                }

                SuperController.LogMessage("Skipping " + filePath + " all nodes look to be filled");
                return true;
            }

            return false;
        }

        public bool StorableInCache(string filePath, int nodeIndex, int totalNodes, string presetType, string storablesString)
        {
           
            string cacheSig = ComputeNodeHash(storablesString);

            //if we have a matching personNode and presetType already return true
            if (cacheFileJSON[filePath]["nodes"][nodeIndex][presetType] != null && cacheFileJSON[filePath]["nodes"][nodeIndex][presetType].Value == cacheSig)
            {
                return true;
            }

            //create cache entry
            bool result = CreateNodeCacheEntry(filePath, nodeIndex, presetType, cacheSig);

            SaveCache();
            return result;
        }

        bool InNodeCache(string cacheSig, string filePath)
        {
            if (cacheFileJSON["nodeHashes"][cacheSig] != null)
            {
                return true;
            }

            cacheFileJSON["nodeHashes"][cacheSig] = filePath;
            return false;
        }

        bool CreateNodeCacheEntry(string filePath, int nodeIndex, string presetType, string cacheSig)
        {
            if (cacheFileJSON[filePath]["nodes"] == null)
            {
                cacheFileJSON[filePath]["nodes"] = new JSONArray();
            }
            if (cacheFileJSON[filePath]["nodes"][nodeIndex][presetType] == null)
            {
                if (InNodeCache(cacheSig, filePath))
                {
                    cacheFileJSON[filePath]["nodes"][nodeIndex][presetType] = cacheSig;
                    return true;
                }
                else
                {
                    cacheFileJSON[filePath]["nodes"][nodeIndex][presetType] = cacheSig;
                }
            }


            return false;
        }

        void LoadCache()
        {
            try
            {
                string jsonString = FileManagerSecure.ReadAllText(Cachepath);
                cacheFileJSON = (JSONClass)JSON.Parse(jsonString);
                UpGradeCache();

            }
            catch (Exception e)
            {
                SuperController.LogMessage("Cache not found or was unreadable. New Cache created.");
                cacheFileJSON = new JSONClass();
                UpGradeCache();
            }

            SaveCache();
        }

        public void SaveCache()
        {
            try
            {
                mVR.SaveJSON(cacheFileJSON, Cachepath);
            }
            catch (Exception e)
            {
                SuperController.LogError("Failed to save Cache " + e);
            }
        }

        void UpGradeCache()
        {
            if (!cacheFileJSON.HasKey("version"))
            {
                SuperController.LogMessage("New cache created. This will cause duplicate entries for any previously created presets.");
                cacheFileJSON = new JSONClass();
                cacheFileJSON["version"].AsInt = 1;
                cacheFileJSON["nodeHashes"] = new JSONClass();

            }
        }

        #region Hash Functions

        string ComputeNodeHash(JSONNode node)
        {
            if (node == null)
                return null;

            string hash = ComputeNodeHash(node.ToString());
            return hash;
        }

        string ComputeNodeHash(string text)
        {
            if (string.IsNullOrEmpty(text))
                return null;

            byte[] raw = System.Text.Encoding.ASCII.GetBytes(text);
            return ComputeNodeHash(raw);
        }

        string ComputeNodeHash(byte[] inputBytes)
        {
            if (inputBytes == null || inputBytes.Length == 0)
                return null;

            using (MD5 md5 = MD5.Create())
            {
                byte[] bytes = md5.ComputeHash(inputBytes);
                StringBuilder builder = new StringBuilder(bytes.Length * 2);
                for (int i = 0; i < bytes.Length; i++)
                {
                    builder.Append(bytes[i].ToString("x2"));
                }
                return builder.ToString();
            }
        }
        #endregion
    }

    class TransformData
    {
        public Vector3 position { get; set; }
        public Quaternion rotation { get; set; }
        public float distance { get; set; }
    }

    class CamPositions
    {
        public string OrigHelpText { get; set; }


        public TransformData VAMCam { get; private set; } = new TransformData();
        public TransformData AppearanceCam { get; private set; } = new TransformData();
        public TransformData ClothingCam { get; private set; } = new TransformData();
        public TransformData SkinCam { get; private set; } = new TransformData();
        public TransformData HairCam { get; private set; } = new TransformData();
        public Color OrigHelpColor { get; set; }
        public TransformData headData { get; private set; } = new TransformData();

        Vector3 deltaP = new Vector3();

        public void UpdateHeadDeltas(Transform transform)
        {
            deltaP = transform.position - headData.position;
        }


        public void SetStartPosition(Transform transform)
        {
            headData.position = transform.position;
            headData.rotation = transform.rotation;
        }


        public void LoadCamPosition(string presetType, Transform target)
        {
            switch (presetType)
            {
                default:
                    target.SetPositionAndRotation(VAMCam.position + deltaP, VAMCam.rotation);
                    return;

                case "AppearancePresets":
                    target.SetPositionAndRotation(AppearanceCam.position + deltaP, AppearanceCam.rotation);
                    return;

                case "ClothingPresets":
                    target.SetPositionAndRotation(ClothingCam.position + deltaP, ClothingCam.rotation);
                    return;

                case "SkinPresets":
                    target.SetPositionAndRotation(SkinCam.position + deltaP, SkinCam.rotation);
                    return;

                case "HairPresets":
                    target.SetPositionAndRotation(HairCam.position + deltaP, HairCam.rotation);
                    return;
            }

        }

        public void StoreCamPosition(string presetType, Transform transform)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    AppearanceCam.position = transform.position;
                    AppearanceCam.rotation = transform.rotation;
                    AppearanceCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "ClothingPresets":
                    ClothingCam.position = transform.position;
                    ClothingCam.rotation = transform.rotation;
                    ClothingCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "SkinPresets":
                    SkinCam.position = transform.position;
                    SkinCam.rotation = transform.rotation;
                    SkinCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "HairPresets":
                    HairCam.position = transform.position;
                    HairCam.rotation = transform.rotation;
                    HairCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                default:
                    VAMCam.position = transform.position;
                    VAMCam.rotation = transform.rotation;
                    return;
            }
        }
    }

    #region Storables
    //pulled out of VAM. Creates a preset string for cache comparison. Does not fullt match VAM output order!!!
    class FakeStoreStorables
    {
        MVRScript mVRScript;
        PresetManager presetManager;

        protected List<PresetManager.Storable> storables;
        protected List<PresetManager.Storable> optionalStorables;
        protected List<PresetManager.Storable> optionalStorables2;
        protected List<PresetManager.Storable> optionalStorables3;
        protected List<PresetManager.Storable> dynamicStorables;
        protected Dictionary<JSONStorable, bool> regularStorables;
        protected Dictionary<string, bool> specificKeyStorables;

        public FakeStoreStorables(MVRScript mVR)
        {
            mVRScript = mVR;
        }

        public string GetStorablesString(Atom containingAtom, string presetType = "AppearancePresets")
        {
            List<MeshVR.PresetManagerControl> pmcs = containingAtom.presetManagerControls;
            MeshVR.PresetManagerControl appearancePresetControl = pmcs.Where(x => x.name == presetType).FirstOrDefault();

            presetManager = appearancePresetControl.GetComponent<PresetManager>();


            JSONClass jsonclass = new JSONClass();
            jsonclass["setUnlistedParamsToDefault"].AsBool = true;
            if (presetManager.conditionalFlagsToStore != null && (presetManager.storeConditionalFlagsAlways || (presetManager.storeConditionalFlagsWhenStoreOptional && presetManager.storeOptionalStorables) || (presetManager.storeConditionalFlagsWhenStoreOptional2 && presetManager.storeOptionalStorables2) || (presetManager.storeConditionalFlagsWhenStoreOptional3 && presetManager.storeOptionalStorables3)))
            {
                foreach (string aKey in presetManager.conditionalFlagsToStore)
                {
                    jsonclass[aKey].AsBool = true;
                }
            }

            RefreshStorables();

            StoreStorables(jsonclass);
            StringBuilder stringBuilder = new StringBuilder(100000);
            jsonclass.ToString(string.Empty, stringBuilder);
            string value = stringBuilder.ToString();

            return value;
        }

        protected void StoreStorablesInList(JSONClass jc, List<PresetManager.Storable> sts)
        {
            if (sts != null)
            {
                foreach (MeshVR.PresetManager.Storable storable in sts)
                {
                    JSONStorable storable2 = storable.storable;
                    if (storable2 != null && (presetManager.ignoreExclude || !storable2.exclude) && storable2.gameObject.activeInHierarchy)
                    {
                        try
                        {
                            JSONClass json = storable2.GetJSON(presetManager.includePhysical, presetManager.includeAppearance, true);
                            if (storable.specificKey != null && storable.specificKey != string.Empty)
                            {
                                JSONClass jsonclass = new JSONClass();
                                foreach (string text in json.Keys)
                                {
                                    if (text == storable.specificKey || text == "id")
                                    {
                                        jsonclass[text] = json[text];
                                    }
                                }
                                jc["storables"].Add(jsonclass);
                            }
                            else
                            {
                                jc["storables"].Add(json);
                            }
                        }
                        catch (Exception ex)
                        {
                            SuperController.LogError(string.Concat(new object[]
                            {
                                "Exception during Preset Store of ",
                                storable2.storeId,
                                ": ",
                                ex
                            }));
                        }
                    }
                }
            }
        }


        protected void RefreshSpecificStorables(PresetManager.SpecificStorable[] spStorables, List<PresetManager.Storable> outputStorables)
        {
            foreach (PresetManager.SpecificStorable specificStorable in spStorables)
            {
                if (specificStorable.specificStorableBucket != null)
                {
                    JSONStorable[] array;
                    if (specificStorable.includeChildren)
                    {
                        array = specificStorable.specificStorableBucket.GetComponentsInChildren<JSONStorable>(true);
                    }
                    else
                    {
                        array = specificStorable.specificStorableBucket.GetComponents<JSONStorable>();
                    }
                    foreach (JSONStorable jsonstorable in array)
                    {
                        if (!(jsonstorable is PresetManagerControl))
                        {
                            if (specificStorable.storeId == string.Empty || specificStorable.storeId == jsonstorable.storeId)
                            {
                                PresetManager.Storable storable = new PresetManager.Storable();
                                storable.storable = jsonstorable;
                                if (specificStorable.specificKey != null && specificStorable.specificKey != string.Empty && !specificKeyStorables.ContainsKey(specificStorable.storeId))
                                {
                                    specificKeyStorables.Add(specificStorable.storeId, true);
                                }
                                storable.specificKey = specificStorable.specificKey;
                                storable.isSpecificKeyAnObject = specificStorable.isSpecificKeyAnObject;
                                outputStorables.Add(storable);
                                if (!regularStorables.ContainsKey(jsonstorable))
                                {
                                    regularStorables.Add(jsonstorable, true);
                                }
                            }
                        }
                    }
                }
            }
        }

        public void RefreshStorables()
        {
            if (Application.isPlaying)
            {
                storables = new List<PresetManager.Storable>();
                optionalStorables = new List<PresetManager.Storable>();
                optionalStorables2 = new List<PresetManager.Storable>();
                optionalStorables3 = new List<PresetManager.Storable>();
                regularStorables = new Dictionary<JSONStorable, bool>();
                specificKeyStorables = new Dictionary<string, bool>();
                if (presetManager.specificStorables != null && presetManager.specificStorables.Length > 0)
                {
                    RefreshSpecificStorables(presetManager.specificStorables, storables);
                }
                if (presetManager.optionalSpecificStorables != null && presetManager.optionalSpecificStorables.Length > 0)
                {
                    RefreshSpecificStorables(presetManager.optionalSpecificStorables, optionalStorables);
                }
                if (presetManager.optionalSpecificStorables2 != null && presetManager.optionalSpecificStorables2.Length > 0)
                {
                    RefreshSpecificStorables(presetManager.optionalSpecificStorables2, optionalStorables2);
                }
                if (presetManager.optionalSpecificStorables3 != null && presetManager.optionalSpecificStorables3.Length > 0)
                {
                    RefreshSpecificStorables(presetManager.optionalSpecificStorables3, optionalStorables3);
                }
                if (presetManager.useTransformAndChildren)
                {
                    JSONStorable[] componentsInChildren = mVRScript.GetComponentsInChildren<JSONStorable>(true);
                    foreach (JSONStorable jsonstorable in componentsInChildren)
                    {
                        if (!regularStorables.ContainsKey(jsonstorable) && !(jsonstorable is PresetManagerControl))
                        {
                            PresetManager.Storable storable = new PresetManager.Storable();
                            storable.storable = jsonstorable;
                            storables.Add(storable);
                            regularStorables.Add(jsonstorable, true);
                        }
                    }
                }
            }
        }
        void RefreshDynamicStorables()
        {
            dynamicStorables = new List<PresetManager.Storable>();
            foreach (Transform transform in presetManager.dynamicStorablesBuckets)
            {
                JSONStorable[] componentsInChildren = transform.GetComponentsInChildren<JSONStorable>(true);
                foreach (JSONStorable jsonstorable in componentsInChildren)
                {
                    if (!regularStorables.ContainsKey(jsonstorable))
                    {
                        PresetManager.Storable storable = new PresetManager.Storable();
                        storable.storable = jsonstorable;
                        dynamicStorables.Add(storable);
                    }
                }
            }
        }
        public void StoreStorables(JSONClass jc)
        {
            if (jc != null)
            {
                RefreshDynamicStorables();
                jc["storables"] = new JSONArray();
                if (presetManager.optionalFirst)
                {
                    if (presetManager.includeOptional && presetManager.storeOptionalStorables)
                    {
                        StoreStorablesInList(jc, optionalStorables);
                    }
                    if (presetManager.includeOptional2 && presetManager.storeOptionalStorables2)
                    {
                        StoreStorablesInList(jc, optionalStorables2);
                    }
                    if (presetManager.includeOptional3 && presetManager.storeOptionalStorables3)
                    {
                        StoreStorablesInList(jc, optionalStorables3);
                    }
                }
                StoreStorablesInList(jc, storables);
                StoreStorablesInList(jc, dynamicStorables);
                if (!presetManager.optionalFirst)
                {
                    if (presetManager.includeOptional && presetManager.storeOptionalStorables)
                    {
                        StoreStorablesInList(jc, optionalStorables);
                    }
                    if (presetManager.includeOptional2 && presetManager.storeOptionalStorables2)
                    {
                        StoreStorablesInList(jc, optionalStorables2);
                    }
                    if (presetManager.includeOptional3 && presetManager.storeOptionalStorables3)
                    {
                        StoreStorablesInList(jc, optionalStorables3);
                    }
                }
            }
        }

    }
    #endregion
}